﻿

/**
 * wisej.web.ext.ImageUpload
 * 
 * Extends the wisej.web.Upload widget to resize an image on the client
 * before uploading it to the server.
 */
qx.Class.define("wisej.web.ext.ImageUpload", {

	extend: wisej.web.Upload,

	properties: {

		/**
		 * Image resize dimensions in pixels: {width, height}.
		 */
		imageSize: { init: null, check: "Map" },

		/**
		 * Image resize percent. When above 0, resizes image by a percent (ie, 50% = .50)
		 */
		imageSizePercent: { init: -1.0, check: "Number" },

		/**
		 * The image format. The default type is image/png.
		 */
		mimeType: { init: "image/png", check: "String" },

		/**
		 * A Number between 0 and 1 indicating image quality if the requested type is image/jpeg or image/webp.
		 */
		quality: { init: 1, check: "Number"}
	},

	statics: {

		/** 
		 * Uploads the files to the server using the submitUrl which may wire the
		 * request to a specific component or to the application instance (id="app").
		 *
		 * @param files {Array} the array of files to package and send.
		 * @param submitUrl {String} the url to submit the data to.
		 * @param filter {String} the allowed file types using this format: "file_extension|audio/*|video/*|image/*|media_type".
		 * @param maxSize {Integer} the maximum size in bytes for each file.
		 * @param callbacks {Map} a map with the following handlers:
		 *
		 *		uploading(fileList[])
		 *		uploaded(fileList[])
		 *		completed(error)
		 *
		 * @returns fileList {Array} the list of file names sent to the server.
		 */
		uploadFiles: function (files, submitUrl, filter, maxSize, callbacks) {

			if (!submitUrl)
				return;

			// normalize the callbacks.
			callbacks = callbacks || {};

			var fileList = [];
			var fileTooLargeNames = [];
			var fileTooLargeSizes = [];
			var formData = new FormData();

			// filter the files first.
			for (var i = 0; i < files.length; i++) {
				var file = files[i];

				// enforce max size.
				if (maxSize > 0 && file.size > maxSize) {

					// save the file too large reference to report the error to the server.
					fileTooLargeNames.push(file.name);
					fileTooLargeSizes.push(file.size);
					continue;
				}

				fileList.push(file.name);
				formData.append('file', file.blob, file.name);
			}

			// don't post if there are no files to post.
			if (fileList.length > 0) {

				// send the data to our handler.
				var xhr = new XMLHttpRequest();
				xhr.open("POST", submitUrl, true);
				xhr.setRequestHeader("X-Requested-With", "XMLHttpRequest");
				xhr.setRequestHeader("X-Wisej-RequestType", "Postback");

				// uploading...
				if (callbacks.uploading)
					callbacks.uploading(fileList, xhr);

				// progress callback.
				xhr.upload.onprogress = callbacks.progress;

				xhr.onreadystatechange = function () {

					if (xhr.readyState == 4) {

						if (xhr.status === 200) {

							// uploaded...
							if (callbacks.uploaded)
								callbacks.uploaded(fileList);

							// completed...
							if (callbacks.completed)
								callbacks.completed();

							// let Wisej process the response from the server.
							Wisej.Core.processResponse.call(Wisej.Core, xhr.responseText);

						} else {

							// completed with error...
							if (callbacks.completed)
								callbacks.completed({ error: "upload", message: xhr.statusText });

						}
					}
				};
				xhr.send(formData);
			}

			// report the list of files that are too large.
			if (fileTooLargeNames.length > 0 && callbacks.completed)
				callbacks.completed({ error: "fileTooLarge", fileNames: fileTooLargeNames, fileSizes: fileTooLargeSizes });

			return fileList;
		},

	},

	members: {

		/**
		 * Invoked when an image has been resized and placed in the new files collection.
		 */
		_onImageResized: function (files, fileIndex, success) {

			// send the files when all the selected files have been processed.
			for (var i = 0; i < files.length; i++) {

				if (files[i].blob == null)
					return;
			}

			this.__sendFiles(files);
		},

		/**
		 * Uploads the resized images.
		 */
		__sendFiles: function (files) {

			// don't post if the session is expired or the submitURL has not
			// been set.
			var submitUrl = this.getSubmitURL();
			if (!submitUrl)
				return;

			if (Wisej.Core.session.expired)
				return;

			if (files == null || files.length == 0)
				return;

			// send files back to this widget.
			// submitUrl += "&v=" + Date.now();
			var me = this;
			this.__currentXhr = null;
			var fileList = wisej.web.ext.ImageUpload.uploadFiles(
				files,
				submitUrl,
				this.getAllowedFileTypes(),
				this.getMaxFileSize(),
				{
					uploading: function (list, xhr) {

						me.__currentXhr = xhr;

						if (me.getShowLoader())
							me.showLoader();
					},

					uploaded: function (fileList) {

						me.__currentXhr = null;

						// fire the "uploaded" event.
						me.fireDataEvent("uploaded", fileList);
					},

					completed: function (error) {

						me.hideLoader();
						me.__currentXhr = null;

						if (error)
							// fire the "error" event.
							me.fireDataEvent("error", error);
					},

					progress: function (evt) {

						me.fireDataEvent("progress", { loaded: evt.loaded, total: evt.total });

					}
				}
			);

			// fire the "valueChanged" event.
			this.fireDataEvent("valueChanged", fileList);

		},

		/**
		 * Uploads the selected files.
		 */
		__onUpload: function (e) {

			// copy the value from the inner upload control, when in design mode.
			if (wisej.web.DesignMode) {
				var textField = this.getChildControl("textfield");
				textField.setValue(this.__upload.value);
				return;
			}

			// don't post if the session is expired or the submitURL has not
			// been set.
			var submitUrl = this.getSubmitURL();
			if (!submitUrl)
				return;

			if (Wisej.Core.session.expired)
				return;

			// read selected files.
			var files = this.getFiles();
			if (files == null || files.length == 0)
				return;

			// transform the files array to our new blob array.
			files = Array.prototype.map.call(files, function (f) {
				return {
					blob: null,
					file: f,
					name: f.name,
					size: f.size
				};
			});

			// read and resize each image.
			for (var i = 0; i < files.length; i++) {

				var reader = new qx.bom.FileReader();
				reader.setUserData("files", files);
				reader.setUserData("fileIndex", i);
				reader.addListenerOnce("load", this.__onFileLoaded, this);
				reader.readAsDataURL(files[i].file);

			}
		},

		/**
		 * Handles the "load" event from the FileReader.
		 */
		__onFileLoaded: function (e) {

			var me = this;
			var reader = e.getTarget();
			var quality = this.getQuality();
			var mimeType = this.getMimeType();
			var imageSize = this.getImageSize();
			var imageSizePercent = this.getImageSizePercent();
			var canvas = document.createElement("canvas");
			var ctx = canvas.getContext("2d");
			var img = new Image();

			var files = reader.getUserData("files");
			var fileIndex = reader.getUserData("fileIndex");

			img.onload = function () {

				canvas.width = img.width;
				canvas.height = img.height;
				ctx.drawImage(img, 0, 0, img.width, img.height);
				canvas.toBlob(function (blob) {

					img.remove();
					reader.dispose();
					files[fileIndex].blob = blob;
					me._onImageResized(files, fileIndex, true);

				}, mimeType, quality);
			}

			img.onerror = function () {
				img.remove();
				reader.dispose();
				files[fileIndex].blob = new Blob();
				me._onImageResized(files, fileIndex, false);
			}

			img.style.top = "-10000px";
			img.style.left = "-10000px";
			img.style.position = "absolute";
			document.body.appendChild(img);
			if (imageSizePercent > 0) {
				img.width = img.width * imageSizePercent;
				img.height = img.height * imageSizePercent;
			}
			else {
				if (imageSize.width)
					img.width = imageSize.width;
				if (imageSize.height)
					img.height = imageSize.height;
			}

			img.src = e.getData().content;
		}
	}

});
